# ERRORS
# Series to be followed = 7XX
BAD_INPUT_DATA_URL = 701
BAD_INPUT_DATA_ALIAS_URL = 702
BAD_INPUT_DATA_ALIAS_MAP = 703
BAD_INPUT_ATTRIBUTE_DATA_TYPE = 704
BAD_INPUT_BAD_DATA_LABEL_ATTRIBUTE = 705
BAD_INPUT_DEPENDENCY_PARSER_NOT_SPECIFIED = 706
